using System;
using System.Collections;
using skmDataStructures.Graph;

namespace DAWalkerProg {
    class DijkstraImpl {
        private Hashtable dist = new Hashtable();
        private Hashtable route = new Hashtable();
        public Graph SoCalMap = new Graph();

        /// <summary>
        /// Retrieves the Node from the passed-in NodeList that has the <i>smallest</i> value in the distance table.
        /// </summary>
        /// <remarks>This method of grabbing the smallest Node gives Dijkstra's Algorithm a running time of
        /// O(<i>n</i><sup>2</sup>), where <i>n</i> is the number of nodes in the graph.  A better approach is to
        /// use a <i>priority queue</i> data structure to store the nodes, rather than an array.  Using a priority queue
        /// will improve Dijkstra's running time to O(E lg <i>n</i>), where E is the number of edges.  This approach is
        /// preferred for sparse graphs.  For more information on this, consult the README included in the download.</remarks>
        private Node GetMin(NodeList nodes) {
            // find the node in nodes with the smallest distance value
            int minDist = Int32.MaxValue;
            Node minNode = null;
            foreach(Node n in nodes) {
                if(((int)dist[n.Key]) <= minDist) {
                    minDist = (int)dist[n.Key];
                    minNode = n;
                }
            }

            return minNode;
        }
        /// <summary>
        /// Initializes the distance and route tables used for Dijkstra's Algorithm.
        /// </summary>
        /// <param name="start">The <b>Key</b> to the source Node.</param>
        private void InitDistRouteTables(string start) {
            // set the initial distance and route for each city in the graph
            foreach(Node n in SoCalMap.Nodes) {
                dist.Add(n.Key, Int32.MaxValue);
                route.Add(n.Key, null);
            }

            // set the initial distance of start to 0
            dist[start] = 0;
        }
        /// <summary>
        /// Relaxes the edge from the Node uCity to vCity.
        /// </summary>
        /// <param name="cost">The distance between uCity and vCity.</param>
        private void Relax(Node uCity, Node vCity, int cost) {
            int distTouCity = (int)dist[uCity.Key];
            int distTovCity = (int)dist[vCity.Key];

            if(distTovCity > distTouCity + cost) {
                // update distance and route
                dist[vCity.Key] = distTouCity + cost;
                route[vCity.Key] = uCity;
            }
        }
        /// <summary>
        /// Runs Dijkstra's Algorithm to find the shortest path.
        /// </summary>
        public int[] ShortestPath(string start, string end) {
            if(start == end) return new int[] { int.Parse(end) };
            dist.Clear();
            route.Clear();
            InitDistRouteTables(start);		// initialize the route & distance tables

            NodeList nodes = SoCalMap.nodes;	// nodes == Q
            Hashtable temp = new Hashtable(nodes.data);

            /**** START DIJKSTRA ****/
            while(nodes.Count > 0) {
                Node u = GetMin(nodes);		// get the minimum node
                nodes.Remove(u);			// remove it from the set Q

                // iterate through all of u's neighbors
                foreach(EdgeToNeighbor edge in u.Neighbors) Relax(u, edge.Neighbor, edge.Cost);		// relax each edge
            }	// repeat until Q is empty
            /**** END DIJKSTRA ****/
            nodes.data = temp;

            Stack traceBackSteps = new Stack();

            Node current = new Node(end, null);
            Node prev = null;
            do {
                prev = current;
                current = (Node)route[current.Key];

                traceBackSteps.Push(prev.Key);
                if(traceBackSteps.Count > 500) return null; //--Couldn't Find Path--The "current != null" doesn't seem to work.  The Relax edges keeps adding a repeating "... > 5001(MapBeforeCR1) > 5101(CR1) > 5001 > 5101..." for when you start in a map with no exit portals.  Something Wrong with Graph? or Something Wrong with Djik?
            } while(current != null && current.Key != start);
            if(end != start) traceBackSteps.Push(start);

            int[] intPath = new int[traceBackSteps.Count];

            for(int i = 0; traceBackSteps.Count > 0; i++) intPath[i] = int.Parse((string)traceBackSteps.Pop());
            if((int)dist[end] < 0) return null;
            else return intPath;
        }
    }
}